@@ -55,6 +55,7 @@
 
 static Package *parse_manifest(Options *op);
 static int install_kernel_module(Options *op,  Package *p);
+static void free_package(Package *p);
 
 
 /*
@@ -116,7 +117,7 @@ int install_from_cwd(Options *op)
     
     /* ask the user to accept the license */
     
-    if (!get_license_acceptance(op)) return FALSE;
+    if (!get_license_acceptance(op)) goto exit_install;
     
     ui_log(op, "Installing NVIDIA driver version %s.", p->version);
 
@@ -125,7 +126,7 @@ int install_from_cwd(Options *op)
      * they really want to overwrite the existing installation
      */
 
-    if (!check_for_existing_driver(op, p)) return FALSE;
+    if (!check_for_existing_driver(op, p)) goto exit_install;
 
     /* run the distro preinstall hook */
 
@@ -138,6 +139,10 @@ int install_from_cwd(Options *op)
     }
     ran_pre_install_hook = TRUE;
 
+    /* fail if the nouveau driver is currently in use */
+
+    if (!check_for_nouveau(op)) goto failed;
+
     /* attempt to build a kernel module for the target kernel */
 
     if (!op->no_kernel_module) {
@@ -224,7 +229,9 @@ int install_from_cwd(Options *op)
 
     /* call the ui to get approval for the list of commands */
     
-    if (!ui_approve_command_list(op, c, "%s", p->description)) return FALSE;
+    if (!ui_approve_command_list(op, c, "%s", p->description)) {
+        goto exit_install;
+    }
     
     /* initialize the backup log file */
 
@@ -290,10 +297,17 @@ int install_from_cwd(Options *op)
         }
     }
     
+    free_package(p);
+
     return TRUE;
     
  failed:
 
+    /*
+     * something bad happened during installation; print an error
+     * message and return FALSE
+     */
+    
     if (op->logging) {
         ui_error(op, "Installation has failed.  Please see the file '%s' "
                  "for details.  You may find suggestions on fixing "
@@ -309,6 +323,18 @@ int install_from_cwd(Options *op)
     if (ran_pre_install_hook)
         run_distro_hook(op, "failed-install");
 
+    /* fall through into exit_install... */
+
+ exit_install:
+
+    /*
+     * we are exiting installation; this can happen for reasons that
+     * do not merit the error message (e.g., the user declined the
+     * license agreement)
+     */
+    
+    free_package(p);
+    
     return FALSE;
 
 } /* install_from_cwd() */
@@ -432,6 +458,8 @@ int add_this_kernel(Options *op)
 
     if (!pack_precompiled_kernel_interface(op, p)) goto failed;
     
+    free_package(p);
+
     return TRUE;
 
  failed:
@@ -439,6 +467,8 @@ int add_this_kernel(Options *op)
     ui_error(op, "Unable to add a precompiled kernel interface for the "
              "running kernel.");
     
+    free_package(p);
+
     return FALSE;
     
 } /* add_this_kernel() */
@@ -479,7 +509,7 @@ static Package *parse_manifest (Options *op)
 {
     char *buf, *c, *flag , *tmpstr;
     int done, n, line;
-    int fd, len = 0;
+    int fd, ret, len = 0;
     struct stat stat_buf, entry_stat_buf;
     Package *p;
     char *manifest = MAP_FAILED, *ptr;
@@ -593,7 +623,10 @@ static Package *parse_manifest (Options *op)
     
     do {
         buf = get_next_line(ptr, &ptr, manifest, len);
-        if ((!buf) || (buf[0] == '\0')) {
+        if (!buf) {
+            done = TRUE;
+        } else if (buf[0] == '\0') {
+            free(buf);
             done = TRUE;
         } else {
             
@@ -609,23 +642,29 @@ static Package *parse_manifest (Options *op)
                 goto fail;
             }
             
+            /* initialize the new entry */
+
+            memset(&p->entries[n], 0, sizeof(PackageEntry));
+
             /* read the file name and permissions */
 
             c = buf;
 
             p->entries[n].file = read_next_word(buf, &c);
+
+            if (!p->entries[n].file) goto invalid_manifest_file;
+
             tmpstr = read_next_word(c, &c);
             
-            /* if any of them were NULL, fail */
-
-            if (!p->entries[n].file || !tmpstr) goto invalid_manifest_file;
+            if (!tmpstr) goto invalid_manifest_file;
             
             /* translate the mode string into an octal mode */
             
-            if (!mode_string_to_mode(op, tmpstr, &p->entries[n].mode)) {
-                goto invalid_manifest_file;
-            }
+            ret = mode_string_to_mode(op, tmpstr, &p->entries[n].mode);
+
             free(tmpstr);
+
+            if (!ret) goto invalid_manifest_file;
             
             /* every file has a type field */
 
@@ -640,8 +679,12 @@ static Package *parse_manifest (Options *op)
                 p->entries[n].flags |= FILE_TYPE_KERNEL_MODULE_CMD;
             else if (strcmp(flag, "OPENGL_HEADER") == 0)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_HEADER;
+            else if (strcmp(flag, "CUDA_HEADER") == 0)
+                p->entries[n].flags |= FILE_TYPE_CUDA_HEADER;
             else if (strcmp(flag, "OPENGL_LIB") == 0)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_LIB;
+            else if (strcmp(flag, "CUDA_LIB") == 0)
+                p->entries[n].flags |= FILE_TYPE_CUDA_LIB;
             else if (strcmp(flag, "LIBGL_LA") == 0)
                 p->entries[n].flags |= FILE_TYPE_LIBGL_LA;
             else if (strcmp(flag, "XLIB_STATIC_LIB") == 0)
@@ -658,6 +701,8 @@ static Package *parse_manifest (Options *op)
                 p->entries[n].flags |= FILE_TYPE_MANPAGE;
             else if (strcmp(flag, "OPENGL_SYMLINK") == 0)
                 p->entries[n].flags |= FILE_TYPE_OPENGL_SYMLINK;
+            else if (strcmp(flag, "CUDA_SYMLINK") == 0)
+                p->entries[n].flags |= FILE_TYPE_CUDA_SYMLINK;
             else if (strcmp(flag, "XLIB_SYMLINK") == 0)
                 p->entries[n].flags |= FILE_TYPE_XLIB_SYMLINK;
             else if (strcmp(flag, "TLS_SYMLINK") == 0)
@@ -676,6 +721,8 @@ static Package *parse_manifest (Options *op)
                 p->entries[n].flags |= FILE_TYPE_XMODULE_SHARED_LIB;
             else if (strcmp(flag, "XMODULE_SYMLINK") == 0)
                 p->entries[n].flags |= FILE_TYPE_XMODULE_SYMLINK;
+            else if (strcmp(flag, "XMODULE_NEWSYM") == 0)
+                p->entries[n].flags |= FILE_TYPE_XMODULE_NEWSYM;
             else {
                 nvfree(flag);
                 goto invalid_manifest_file;
@@ -728,9 +775,9 @@ static Package *parse_manifest (Options *op)
                 p->entries[n].path = NULL;
             }
             
-            /* symlinks have a target */
+            /* symlinks and newsyms have a target */
 
-            if (p->entries[n].flags & FILE_TYPE_SYMLINK) {
+            if (p->entries[n].flags & FILE_TYPE_HAVE_TARGET) {
                 p->entries[n].target = read_next_word(c, &c);
                 if (!p->entries[n].target) goto invalid_manifest_file;
             } else {
@@ -787,8 +834,7 @@ static Package *parse_manifest (Options *op)
     goto fail;
 
  fail:
-    if (p && p->entries) free(p->entries);
-    if (p) free(p);
+    free_package(p);
     if (manifest != MAP_FAILED) munmap(manifest, len);
     if (fd != -1) close(fd);
     return NULL;
@@ -839,3 +885,56 @@ void add_package_entry(Package *p,
 
 } /* add_package_entry() */
 
+
+
+/*
+ * free_package() - free the Package data structure
+ */
+
+static void free_package(Package *p)
+{
+    int i;
+
+    if (!p) return;
+    
+    nvfree(p->description);
+    nvfree(p->version);
+    nvfree(p->kernel_module_filename);
+    nvfree(p->kernel_interface_filename);
+    nvfree(p->kernel_module_name);
+    
+    if (p->bad_modules) {
+        for (i = 0; p->bad_modules[i]; i++) {
+            nvfree(p->bad_modules[i]);
+        }
+        nvfree((char *) p->bad_modules);
+    }
+
+    if (p->bad_module_filenames) {
+        for (i = 0; p->bad_module_filenames[i]; i++) {
+            nvfree(p->bad_module_filenames[i]);
+        }
+        nvfree((char *) p->bad_module_filenames);
+    }
+    
+    nvfree(p->kernel_module_build_directory);
+    
+    nvfree(p->precompiled_kernel_interface_directory);
+    
+    for (i = 0; i < p->num_entries; i++) {
+        nvfree(p->entries[i].file);
+        nvfree(p->entries[i].path);
+        nvfree(p->entries[i].target);
+        nvfree(p->entries[i].dst);
+
+        /*
+         * Note: p->entries[i].name just points into
+         * p->entries[i].file, so don't free p->entries[i].name
+         */
+    }
+
+    nvfree((char *) p->entries);
+
+    nvfree((char *) p);
+    
+} /* free_package() */
